#ifndef FreenectQGLWidget_HXX
#define FreenectQGLWidget_HXX

#include "freenectrgbdevice.hxx"
#include <QtOpenGL>

class FreenectQGLWidget : public QGLWidget
{
	public:
		//Create a new Widget
		FreenectQGLWidget( QWidget *parent, int update_interval,
					  FreenectRGBDevice* dev) : QGLWidget( parent)
		{
			m_device = dev;
			m_rgb	=  new uint8_t[m_device->getBufferSize()];
			m_depth	=  new uint8_t[m_device->getBufferSize()];
			
			//Create the timer
			m_timer = new QTimer(this);
			connect(m_timer, SIGNAL(timeout()), this, SLOT(update()));
			m_timer->start(update_interval);
		}
	
		//Destroy an existing Widget
		~FreenectQGLWidget()
		{
			m_timer->stop();
			delete m_timer;
			delete[] m_rgb;
			delete[] m_depth; 
		}
	
	protected:
	
		//Initialize the GL system for this widget
		void initializeGL()
		{
			glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
			
			glClearDepth(1.0);
			glDepthFunc(GL_LESS);
			glDisable(GL_DEPTH_TEST);
			glEnable(GL_BLEND);
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glShadeModel(GL_SMOOTH);
			glGenTextures(1, &m_gl_depth_tex);
			glBindTexture(GL_TEXTURE_2D, m_gl_depth_tex);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glGenTextures(1, &m_gl_rgb_tex);
			glBindTexture(GL_TEXTURE_2D, m_gl_rgb_tex);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			
			resizeGL(this->width(), this->height());
		}
		
		//Handle a resizing of the widget
		// @param width		the new width
		// @param height	the new height of the widget
		void resizeGL( int width, int height )
		{
			glViewport(0,0,width,height);
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			glOrtho (0, 1280, 480, 0, -10.0f, 10.0f);
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
		}
		
		//Draw both, the depth field and the rgb image using OpenGL 
		void paintGL()
		{
			m_device->getDepth(m_depth);
			m_device->getRGB(m_rgb);
						
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glLoadIdentity();
			
			glEnable(GL_TEXTURE_2D);
			
			glBindTexture(GL_TEXTURE_2D, m_gl_depth_tex);
			glTexImage2D(GL_TEXTURE_2D, 0, 3, 640, 480, 0, GL_RGB, GL_UNSIGNED_BYTE, m_depth);
			
			glBegin(GL_TRIANGLE_FAN);
			glColor4f(255.0f, 255.0f, 255.0f, 255.0f);
			glTexCoord2f(0, 0); glVertex3f(0,0,0);
			glTexCoord2f(1, 0); glVertex3f(640,0,0);
			glTexCoord2f(1, 1); glVertex3f(640,480,0);
			glTexCoord2f(0, 1); glVertex3f(0,480,0);
			glEnd();
			glBindTexture(GL_TEXTURE_2D, m_gl_rgb_tex);
			
			if (m_device->getVideoFormat() == FREENECT_VIDEO_RGB || m_device->getVideoFormat() == FREENECT_VIDEO_YUV_RGB)
				glTexImage2D(GL_TEXTURE_2D, 0, 3, 640, 480, 0, GL_RGB, GL_UNSIGNED_BYTE, m_rgb);
			else
				glTexImage2D(GL_TEXTURE_2D, 0, 1, 640, 480, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, m_rgb);
			
			glBegin(GL_TRIANGLE_FAN);
			glColor4f(255.0f, 255.0f, 255.0f, 255.0f);
			glTexCoord2f(0, 0); glVertex3f(640,0,0);
			glTexCoord2f(1, 0); glVertex3f(1280,0,0);
			glTexCoord2f(1, 1); glVertex3f(1280,480,0);
			glTexCoord2f(0, 1); glVertex3f(640,480,0);
			glEnd();
		}
	
		//Handle Keys which are pressed by the user
		void keyPressEvent(QKeyEvent *event)
		{
			int	key	= event->key();
			freenect_video_format	requested_format	=	m_device->getVideoFormat();
			
			switch (key){
				case Qt::Key_0:
					m_device->setLed(LED_OFF);
					break;
				case Qt::Key_1:
					m_device->setLed(LED_GREEN);
					break;
				case Qt::Key_2:
					
					m_device->setLed(LED_RED);
					break;
					
				case Qt::Key_3:
					m_device->setLed(LED_YELLOW);
					break;
					
				case Qt::Key_4:
					m_device->setLed(LED_BLINK_GREEN);
					break;
					
				case Qt::Key_5:
					// 5 is the same as 4
					m_device->setLed(LED_BLINK_GREEN);
					break;
					
				case Qt::Key_6:
					m_device->setLed(LED_BLINK_RED_YELLOW);
					break;
			
				case Qt::Key_F:
					if (requested_format == FREENECT_VIDEO_IR_8BIT) {
						requested_format = FREENECT_VIDEO_RGB;
					} else if (requested_format == FREENECT_VIDEO_RGB){
						requested_format = FREENECT_VIDEO_YUV_RGB;
					} else {
						requested_format = FREENECT_VIDEO_IR_8BIT;
					}
					m_device->setVideoFormat(requested_format);
					break;
				
				case Qt::Key_W:
					m_target_angle++;
					if (m_target_angle > 30) {
						m_target_angle = 30;
					}
					break;
					
				case Qt::Key_S:
				case Qt::Key_D:
					m_target_angle = 0;
					break;
					
				case Qt::Key_X:
						m_target_angle--;
					if (m_target_angle < -30) {
						m_target_angle = -30;
					}
					break;
					
				case Qt::Key_E:
					m_target_angle = 10;
					break;
					
				case Qt::Key_C:
					m_target_angle = -10;
					break;
			}
			m_device->setTiltDegrees(m_target_angle);
		}
	
	private:		
		int	m_target_angle;	//The target angle of the freenect tilt:
							//This is needed because of the noisy m_device->getState().m_code
							//data about the current tilt position.
		
		FreenectRGBDevice*  m_device; //The freenect device itself
		
		GLuint m_gl_depth_tex;	//The GL-Texture for the depth field
		GLuint m_gl_rgb_tex;	//The GL-Texture for the rgb image
		
		uint8_t* m_depth;		//raw pointer for depth field
		uint8_t* m_rgb;			//raw pointer for rgb image
	
		QTimer *m_timer;		//Timer to update the scene (by calling paintGL)
};

#endif